#include "calculator.h"

Calculator::Calculator(QWidget* pwgt) : QWidget(pwgt)
{
    m_strDisplay = "";
    m_stk.clear();

    constructRegular();
    constructEngineer();
    constructTopLayout();

    mainWidget = new QWidget(this);
    regularMode->setChecked(true);

    showRegularCalc();
}


void Calculator::constructTopLayout()
{
    m_plcd = new QLCDNumber(12, this);
    m_plcd->setSegmentStyle(QLCDNumber::Flat);
    m_plcd->setMinimumSize(150, 50);
    m_plcd->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);

    regularMode = new QRadioButton("Обычный", this);
    engineerMode = new QRadioButton("Инженерный", this);
    regularMode->setChecked(true);
    engineerMode->setChecked(false);

    modesLayout = new QGridLayout(this);
    modesLayout->addWidget(regularMode, 0, 0);
    modesLayout->addWidget(engineerMode, 1, 0);
    modesWidget = new QWidget(this);
    modesWidget->setLayout(modesLayout);

    ptopLayout = new QGridLayout(this);
    ptopLayout->addWidget(m_plcd, 0, 0, 1, 5);
    ptopLayout->addWidget(modesWidget, 1, 0, 1, 4);
    ptopLayout->addWidget(createButton("CE"), 1, 4);

    topWidget = new QWidget(this);
    topWidget->setLayout(ptopLayout);

    connect(regularMode, SIGNAL(toggled(bool)), SLOT(slotChangeMode()));
}


void Calculator::constructRegular()
{
    QString aButtons[4][5] = {{"7", "8", "9", "/", "sqrt(x)"},
                              {"4", "5", "6", "*", "(1/x)"},
                              {"1", "2", "3", "-", "="},
                              {"0", ".", "+"}};

    regularCalcLayout = new QGridLayout(this);

    for (int i = 0; i < 2; ++i)
    {
        for (int j = 0; j < 5; ++j)
        {
            regularCalcLayout->addWidget(createButton(aButtons[i][j]), i, j);
        }
    }

    for (int j = 0; j < 4; ++j)
    {
        regularCalcLayout->addWidget(createButton(aButtons[2][j]), 4, j);
    }

    QPushButton *equalBtn = createButton(aButtons[2][4]);
    equalBtn->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    regularCalcLayout->addWidget(equalBtn, 4, 4, 2, 1);
    regularCalcLayout->addWidget(createButton(aButtons[3][0]), 5, 0, 1, 2);
    regularCalcLayout->addWidget(createButton(aButtons[3][1]), 5, 2);
    regularCalcLayout->addWidget(createButton(aButtons[3][2]), 5, 3);

    regularCalcWidget = new QWidget(this);
    regularCalcWidget->setLayout(regularCalcLayout);
}


void Calculator::constructEngineer()
{
    QString engButtons[4][4] = {{"sinh", "sin", "exp", "x\u00B2"},
                                {"cosh", "cos", "ln", "x\u00B3"},
                                {"tanh", "tan", "lg", "x\u207F"},
                                {"n!", "\u03C0", "\u00B3\u221Ax", "\u207F\u221Ax"}};

    engineerCalcLayout = new QGridLayout(this);
    for(int i = 0; i < 4; ++i)
    {
        for(int j = 0; j < 4; ++j)
        {
            engineerCalcLayout->addWidget(createButton(engButtons[i][j]), i, j);
        }
    }


    engineerCalcWidget = new QWidget(this);
    engineerCalcWidget->setLayout(engineerCalcLayout);
}


void Calculator::showRegularCalc()
{
    mainLayout = new QGridLayout(this);
    engineerCalcWidget->hide();
    mainLayout->addWidget(topWidget, 0, 0, 2, 5);
    mainLayout->addWidget(regularCalcWidget, 2, 0, 4, 5);
    setLayout(mainLayout);
    setWindowTitle(tr("Regular calculator"));
}


void Calculator::showEngineerCalc()
{
    mainLayout = new QGridLayout(this);
    mainLayout->addWidget(topWidget, 0, 0, 2, 9);
    mainLayout->addWidget(regularCalcWidget, 2, 4, 4, 5);
    mainLayout->addWidget(engineerCalcWidget, 2, 0, 4, 4);
    engineerCalcWidget->show();
    setLayout(mainLayout);
    setWindowTitle(tr("Engineer calculator"));
}


void Calculator::slotChangeMode()
{
    if (regularMode->isChecked())
    {
        delete mainLayout;
        showRegularCalc();
    }
    else
    {
        delete mainLayout;
        showEngineerCalc();
    }

}


QPushButton* Calculator::createButton(const QString &str)
{
    QPushButton* pcmd = new QPushButton(str);
    pcmd->setMinimumSize(40, 40);
    pcmd->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
    connect(pcmd, SIGNAL(clicked()), SLOT(slotButtonClicked()));
    return pcmd;
}


void Calculator::slotButtonClicked()
{
    QString str = ((QPushButton*)sender())->text();
    QVector<QString> bin_ops = {"+", "-", "*", "/", "x\u207F", "\u207F\u221Ax"};


    if (str == "CE")
    {
        m_plcd->display("");
        m_strDisplay = "";
        m_stk.clear();
    }
    else if ((str.contains(QRegularExpression("^\\d"))) || ((str == ".") && (m_strDisplay != ""))) //construct the number
    {
        m_strDisplay += str;
        m_plcd->display(m_strDisplay.toDouble());
    }
    else if (str == "\u03C0")
    {
        m_strDisplay.setNum(M_PI);
        m_plcd->display(m_strDisplay);
    }
    else if ((str == "=") && (m_stk.size() >= 2) && (m_strDisplay != "")) //if equation sign has been pressed
    {
        m_stk.push(QString().setNum(m_plcd->value()));
        if (calculate() == 1)
        {
            m_strDisplay = "";
            m_stk.clear();
            return;
        }
        else
        {
            m_strDisplay = QString().setNum(m_plcd->value());
        }
    }
    else //sign of some operation has been pressed
    {
        if (std::find(bin_ops.begin(), bin_ops.end(), str) != bin_ops.end()) //if chosen operation is binary
        {
            if (m_strDisplay != "") //if there is number in the input window
            {
                m_stk.push(QString().setNum(m_plcd->value()));
            }

            if(m_stk.size() == 3)
            {
                if (calculate() == 0) //success calculation
                {
                    m_stk.push(QString().setNum(m_plcd->value()));
                    m_stk.push(str);
                }
                else
                {
                    m_stk.clear();
                    m_strDisplay = "";
                    return;
                }
            }
            else if (m_stk.size() == 2) //change operation
            {
                m_stk.pop();
                m_stk.push(str);
            }
            else if (m_stk.size() == 1)
            {
                m_stk.push(str);
            }

            m_strDisplay = "";
        }
        else
        {
            if (str == "sqrt(x)")
            {
                if (m_plcd->value() >= 0) m_plcd->display(sqrt(m_plcd->value()));
                else
                {
                    m_plcd->display("Error");
                    m_stk.clear();
                    m_strDisplay = "";
                    return;
                }
            }
            else if (str == "(1/x)")
            {
                if (m_plcd->value() != 0.0) m_plcd->display(1.0 / m_plcd->value());
                else
                {
                    m_plcd->display("Error");
                    m_stk.clear();
                    m_strDisplay = "";
                    return;
                }
            }
            else if (str == "sinh")
            {
                m_plcd->display(sinh(m_plcd->value()));
            }
            else if (str == "cosh")
            {
                m_plcd->display(cosh(m_plcd->value()));
            }
            else if (str == "tanh")
            {
                m_plcd->display(tanh(m_plcd->value()));
            }
            else if (str == "sin")
            {
                m_plcd->display(sin(m_plcd->value()));
            }
            else if (str == "cos")
            {
                m_plcd->display(cos(m_plcd->value()));
            }
            else if (str == "tan")
            {
                m_plcd->display(tan(m_plcd->value()));
            }
            else if (str == "exp")
            {
                m_plcd->display(exp(m_plcd->value()));
            }
            else if (str == "ln")
            {
                if (m_plcd->value() > 0) m_plcd->display(log(m_plcd->value()));
                else
                {
                    m_plcd->display("Error");
                    m_stk.clear();
                    m_strDisplay = "";
                    return;
                }
            }
            else if (str == "lg")
            {
                if (m_plcd->value() > 0) m_plcd->display(log10(m_plcd->value()));
                else
                {
                    m_plcd->display("Error");
                    m_stk.clear();
                    m_strDisplay = "";
                    return;
                }
            }
            else if (str == "x\u00B2")
            {
                m_plcd->display(m_plcd->value() * m_plcd->value());
            }
            else if (str == "x\u00B3")
            {
                m_plcd->display(m_plcd->value() * m_plcd->value() * m_plcd->value());
            }
            else if (str == "\u00B3\u221Ax")
            {
                m_plcd->display(pow(m_plcd->value(), 1.0 / 3));
            }
            else if (str == "n!")
            {
                if (m_plcd->value() >= 0) m_plcd->display(factorial(m_plcd->value()));
                else
                {
                    m_plcd->display("Error");
                    m_stk.clear();
                    m_strDisplay = "";
                    return;
                }
            }
            //m_stk.push(QString().setNum(m_plcd->value()));
            m_strDisplay = QString().setNum(m_plcd->value());
        }
    }
}


int Calculator::factorial(int num)
{
    if (num <= 1) return 1;
    else return num * factorial(num - 1);
}


int Calculator::calculate()
{
    qreal foperand2 = m_stk.pop().toFloat();
    QString strOperation = m_stk.pop();
    qreal fResult = 0;
    qreal foperand1 = m_stk.pop().toFloat();

    if(strOperation == "+")
    {
        fResult = foperand1 + foperand2;
    }
    else if (strOperation == "-")
    {
        fResult = foperand1 - foperand2;
    }
    else if (strOperation == "*")
    {
        fResult = foperand1 * foperand2;
    }
    else if (strOperation == "/")
    {
        if (foperand2 != 0.0)
        {
            fResult = foperand1 / foperand2;
        }
        else
        {
            m_plcd->display("Error");
            return 1;
        }
    }
    else if (strOperation == "x\u207F")
    {
        if (foperand1 == 0.0) fResult = 0.0;
        else
        {
            fResult = pow(foperand1, foperand2);
        }
    }
    else if (strOperation == "\u207F\u221Ax")
    {
        if (foperand1 == 0.0) fResult = 0.0;
        else
        {
            fResult = pow(foperand1, 1.0 / foperand2);
        }
    }

    m_plcd->display(fResult);
    return 0;
}
